
RELEASE_VERSION='0.3.1087709269'
BUILD_DATE='Tue Nov 28 12:34:06 UTC 2023'
BUILD_MACHINE_INFO='Linux runner-bhjdzkxh-project-28359933-concurrent-0hkpbr 5.4.235-144.344.amzn2.x86_64 #1 SMP Sun Mar 12 12:50:22 UTC 2023 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='5632774470'
GIT_COMMIT_SHA='709070bb01ac7755501ef7774537fd0e217af3bd'
GIT_BRANCH='main'
